﻿/**
 * Authors: http://d.hatena.ne.jp/ultraist/
 */
module common.configuration;

/**
 *  設定など
 *
 */

private import std.socket;
private import std.path;

private import pes.net.util;
private import common.keytable;
private import common.nodetable;

class Configuration
{
protected:
	const ushort	DEFAULT_FILE_PORT 		= 3129;
	const ushort	DEFAULT_UI_PORT			= 46490;
	
	char[]		m_noderef					= "Noderef.txt";		// 初期ノードファイル
	char[]		m_cache_folder				= "cache";				// キャッシュフォルダ
	char[]		m_cache_index				= "index.db";			// キャッシュインデックス
	char[]		m_down_folder				= "down";				// ダウンロードフォルダ
	char[]		m_up_folder					= "up";					// アップロードフォルダ
	char[]		m_stt						= "poeny.stt";			// 設定ファイル
	char[]		m_download_list				= "download.stt";		// ダウンロードリスト
	uint		m_max_key					= 30000	;				// 保持するキー数
	uint		m_address;											// 自ノードのアドレス
	ushort		m_file_port					= DEFAULT_FILE_PORT;	// リスナーポート
	ushort		m_ui_port					= DEFAULT_UI_PORT;		// UIポート
	uint		m_speed						= 120;					// 自己申告回線速度
	uint		m_max_node					= 600;					// 保持するノード情報数
	uint		m_max_upstream_connection	= 2;					// 限界検索リンク(UP)
	uint		m_max_downstream_connection	= 5;					// 限界検索リンク(DOWN);
	uint		m_max_data_connection		= 15;					// 限界転送リンク
	bool		m_is_port0					= false;				// ポート0
	uint		m_timeout					= 3;					// コネクションタイムアウト(s)
	uint		m_timeout_download			= 10;					// コネクションタイムアウト(s)ダウンロード時
	uint		m_encoder_max				= 1;					// エンコーダ最大数
	uint		m_decoder_max				= 1;					// デコーダー最大数
	
	version(POEN_TROPY) {
		uint		m_uploader_max				= 5;					// アップローダー最大数
		uint		m_downloader_max			= 4;					// ダウンローダー最大数
	} else {
		uint		m_uploader_max				= 3;					// アップローダー最大数
		uint		m_downloader_max			= 3;					// ダウンローダー最大数
	}
	char[]		m_trip						= "";
	
	// 今のところ固定
	char[][3]	m_clustering_key	= [
//		"合法ファイルで清く正しいP2P", 
//		"ny漫画大王",
//		"ヤムチャ",
		"【もせ】音楽情報交換スレ【アルバム】",
		"自主制作映画 ",
		"合法",
	];
	
	// 初期化
	void initialize()
	{
		// 設定
		m_noderef 					= m_noderef.init;
		m_file_port					= m_file_port.init;
		m_ui_port					= m_ui_port.init;
		m_speed						= m_speed.init;
		m_max_node					= m_max_node.init;
		m_max_upstream_connection	= m_max_upstream_connection.init;
		m_max_downstream_connection	= m_max_downstream_connection.init;
		m_is_port0					= m_is_port0.init;
		m_max_data_connection		= m_max_data_connection.init;
		m_max_key					= m_max_key.init;
		m_address					= getSystemIPAddress().addr;
		m_cache_folder				= m_cache_folder.init;
		m_down_folder				= m_down_folder.init;
		m_up_folder					= m_up_folder.init;
		m_timeout					= m_timeout.init;
		m_trip						= m_trip.init;
	}
	
public:
	this() 
	{
		initialize();
	}
	
	void reset()
	{
		initialize();
	}
	
	void timeout(uint sec)
	{
		m_timeout = sec;
	}
	
	void noderef(char[] file_path)
	{
		m_noderef = file_path.dup;
	}
	
	void speed(uint conn_speed)
	{
		m_speed = conn_speed;
	}
	
	void nodeMax(uint max_node)
	{
		m_max_node = max_node;
	}
	
	void fileSharingPort(ushort port)
	{
		m_file_port = port;
	}
	
	void uiPort(ushort port)
	{
		m_ui_port = port;
	}
	
	void upstreamConnectionMax(uint max)
	{
		m_max_upstream_connection = max;
	}
	
	void downstreamConnectionMax(uint max)
	{
		m_max_downstream_connection = max;
	}
	
	void isPort0(bool v)
	{
		m_is_port0 = v;
	}
	
	void clusteringKey(int index, char[] word)
	{
		m_clustering_key[index] = word.dup;
	}
	
	void clusteringKey(char[][] keys)
	in {
		assert(keys.length == m_clustering_key.length);
	}
	body {
		for (int i = 0; i < m_clustering_key.length; ++i) {
			m_clustering_key[i] = keys[i];
		}
	}
	
	void dataConnectionMax(uint max)
	{
		m_max_data_connection = max;
	}
	
	void keyMax(uint max)
	{
		m_max_key = max;
	}
	
	void cacheFolder(char[] folder)
	{
		m_cache_folder = folder.dup;
	}
	
	void downloadFolder(char[] folder)
	{
		m_down_folder = folder.dup;
	}
	
	void uploadFolder(char[] folder)
	{
		m_up_folder = folder.dup;
	}
	
	void trip(char[] trp)
	{
		m_trip = trp;
	}
	
	void encoderMax(uint max)
	{
		m_encoder_max = max;
	}
	
	void decoderMax(uint max)
	{
		m_decoder_max = max;
	}
	
	void uploaderMax(uint max)
	{
		m_uploader_max = max;
	}
	
	void downloaderMax(uint max)
	{
		m_downloader_max = max;
	}
	
	
	// getter
	
	uint timeout()
	{
		return m_timeout;
	}
	
	char[] noderef()
	{
		return m_noderef;
	}
	
	ushort fileSharingPort()
	{
		return m_file_port;
	}
	
	ushort uiPort()
	{
		return m_ui_port;
	}
	
	int speed()
	{
		return m_speed;
	}
	
	uint nodeMax()
	{
		return m_max_node;
	}
	
	uint upstreamConnectionMax()
	{
		return m_max_upstream_connection;
	}
	
	uint downstreamConnectionMax()
	{
		return m_max_downstream_connection;
	}
	
	bool isPort0()
	{
		return m_is_port0;
	}
	
	char[] clusteringKey(int index)
	{
		return m_clustering_key[index];
	}
	
	char[][] clusteringKey()
	{
		return m_clustering_key.dup;
	}
	
	uint dataConnectionMax()
	{
		return m_max_data_connection;
	}
	
	uint keyMax()
	{
		return m_max_key;
	}
	
	uint address()
	{
		return m_address;
	}
	
	char[] cacheFolder()
	{
		return m_cache_folder;
	}
	
	char[] downloadFolder()
	{
		return m_down_folder;
	}
	
	char[] uploadFolder()
	{
		return m_up_folder;
	}
	
	char[] cacheIndex()
	{
		return m_cache_folder ~ std.path.sep ~ m_cache_index;
	}
	
	char[] trip()
	{
		return m_trip;
	}
	
	uint encoderMax()
	{
		return m_encoder_max;
	}
	
	uint decoderMax()
	{
		return m_decoder_max;
	}
	uint uploaderMax()
	{
		return m_uploader_max;
	}
	
	uint downloaderMax()
	{
		return m_downloader_max;
	}
	
	uint downloadTimeout()
	{
		return m_timeout_download;
	}
	
	char[] stt()
	{
		return m_stt;
	}
	
	char[] downloadList()
	{
		return m_download_list;
	}
}

