﻿/**
 * Authors: http://d.hatena.ne.jp/ultraist/
 */
module common.task;

// タスク云々

private import std.c.stdlib; // EXIT_*
private import std.socket;
private import std.thread;
private import std.stream;
private import std.date;
private import std.string;
private import std.stdio;

private import pes.system.all;
private import pes.net.winny.all;
private import pes.arraybox;
private import pes.prompt;

private import common.environment;
private import common.configuration;

private import pes.prompt;

import pes.prompt;

/* タスクメッセージキュー */

enum MailId: mailbox
{
	UI_REQUEST,			//
	UI_RESPONSE,		
	UI_EXCEPTION,		
	CONTROL_REQUEST,	
	CONTROL_RESPONSE,	
	NODE_REQUEST,		
	NODE_RESPONSE,		
	LINK_REQUEST,
	LINK_RESPONSE,
	CONNECTION_REQUEST,
	CONNECTION_RESPONSE,
	QUERY_REQUEST,		
	QUERY_RESPONSE,		
	KEY_REQUEST,		
	KEY_RESPONSE,		
	DOWNLOAD_REQUEST,	
	DOWNLOAD_RESPONSE,	
	UPLOAD_REQUEST,		
	UPLOAD_RESPONSE,	
	TIMER_REQUEST,		
	TIMER_RESPONSE,		
	
	ANONYMOUS				= mailbox.max,
}

static MailId[] g_mail_ids = [
	MailId.UI_REQUEST,
	MailId.UI_RESPONSE,
	MailId.UI_EXCEPTION,
	MailId.CONTROL_REQUEST,
	MailId.CONTROL_RESPONSE,
	MailId.NODE_REQUEST,
	MailId.NODE_RESPONSE,
	MailId.LINK_REQUEST,
	MailId.LINK_RESPONSE,
	MailId.CONNECTION_REQUEST,
	MailId.CONNECTION_RESPONSE,
	MailId.QUERY_REQUEST,
	MailId.QUERY_RESPONSE,
	MailId.KEY_REQUEST,
	MailId.KEY_RESPONSE,
	MailId.DOWNLOAD_REQUEST,
	MailId.DOWNLOAD_RESPONSE,
	MailId.UPLOAD_REQUEST,
	MailId.UPLOAD_RESPONSE,
	MailId.TIMER_REQUEST,
	MailId.TIMER_RESPONSE,
	
	MailId.ANONYMOUS
];

enum EventId: eventflag
{
	TASK_START				= 0x00000001,	// タスク起動用(TaskId参照)
	RUNNING,								// うんぬん起動中(FunctionId参照)
}

enum TaskId: uint
{
	UI						= 0x00000001,
	CONTROL					= 0x00000002,
	NODE					= 0x00000004,
	QUERY					= 0x00000008,
	KEY						= 0x00000010,
	TIMER					= 0x00000020,
	DOWNLOAD				= 0x00000040,
	UPLOAD					= 0x00000080,
	LINK					= 0x00000100,
	CONNECTION				= 0x00000200,
	//...
	ALL						= 0xFFFFFFFF,
}

enum FunctionId: uint
{
	LINK_RUN			= 0x00000001,	// LinkTask.makeFormation 実行イベント
	LINK_INTERRUPT		= 0x00000002,	// LinkTask.makeFormation 中断イベント
}

static EventId[] g_event_ids = [
	EventId.TASK_START,
	EventId.RUNNING,
];

// セマフォID
enum SemaphoreId: semaphore
{
	KEY			= 0,
	NODE,
	UPSTREAM,
	DOWNSTREAM,
	UL_DATA,
	DL_DATA,
	UPLOADER,
	DOWNLOADER,
	ENCODER,
	DECODER,
	DOWNLOAD_ITEM,
	CACHE_INDEX,
}

SemaphoreId[] g_semaphore_ids =
[
	SemaphoreId.KEY,
	SemaphoreId.NODE,
	SemaphoreId.UPSTREAM,
	SemaphoreId.DOWNSTREAM,
	SemaphoreId.UL_DATA,
	SemaphoreId.DL_DATA,
	SemaphoreId.UPLOADER,
	SemaphoreId.DOWNLOADER,
	SemaphoreId.ENCODER,
	SemaphoreId.DECODER,
	SemaphoreId.DOWNLOAD_ITEM,
	SemaphoreId.CACHE_INDEX,
];

/* 単純処理結果 */
enum RequestStatus: int
{
	OK = 1,
	NG
}


/* メッセージID */
enum MessageId: uint
{
	EXCEPTION,						// 例外通知 (未使用)
	ROUTING,						// ルーティング
	WARNING,						// 警告(未使用)
	RECEIVE_QUERY,					// クエリ受信
	RECEIVE_DIFFUSION_QUERY,		// 拡散クエリ受信
	RECEIVE_COND_DIFFUSION_QUERY,	// 拡散クエリ受信
	ADD_KEY,						// キー追加
	REBOOT,							// タスクリブート
	TERMINATE,						// タスク終了
	ADD_SEARCH_WORD,				// 検索登録
	REMOVE_SEARCH_WORD,				// 検索削除
	SEND_COMMAND,					// コマンド送信
	CLOCK_TICK,						// 1秒間隔のタイマー割り込み
	LINK_CONNECT,					// リンク
	ADD_NODE,						// ノード追加
	REFRESH_NODE,					// ノードリフレッシュ
	ADD_DOWNLOAD,					// ダウンロード追加
	SET_TIMER,						// タイマーセット
	DECODE_CACHE,					// 強制キャッシュ変換
	REMOVE_DOWNLOAD,				// ダウンロード削除(キャンセル)
	REMOVE_CACHE,					// キャッシュ削除
}

alias ArrayBox!(NyKeyInformation)	KeyInformationArray;


/**
 * タスクのベースクラス
 * 
 */
class TaskBase: Task
{
protected:
	enum ReturnCode: int
	{
		SUCCESS		= EXIT_SUCCESS,	//< 成功(本当は成功なんてない)
		FAILURE		= EXIT_FAILURE,	//< 失敗,なんらかエラー(controlは例外で終わる)
		CONTINUE	= 0x0051,		//< もう一度やりたい
		TERMINATE	= 0x5100,		//< 終了します
	}
	
	mailbox				m_response_mailbox;
	Environment			s_environ;
	Thread				m_thread;
	
	abstract void 		initialize();	//< 初期化処理
	abstract void 		terminate();	//< 終了処理
	abstract ReturnCode control();		//< やりたいことをやる
	abstract TaskId 	taskId();		//< 自分のタスクID(bit)を返す
	
	
	override int run()
	{
		int result;
		
		log(std.string.format("start %08X", Thread.getThis.id));
		try {
			result = mainLoop();
		} catch (Object e) {
			result = EXIT_FAILURE;
			errorLog("root Exception: ".dup ~ e.toString);
		}
		
		return result;
	}
	
	
	/** タスクメインループ	*/
	override int mainLoop()
	{
		ReturnCode result = ReturnCode.FAILURE;
		
		m_thread = Thread.getThis();
		
		// 起動イベント待ち
		waitForStartEvent(EventId.TASK_START, taskId());
		
		try {
			initialize();
		} catch (Exception e) {
			errorLog("initialize Exception: ".dup ~ e.toString);
			return cast(int)ReturnCode.FAILURE; // 異常終了
		}
		
		// 起動通知としてイベントを終了させる
		endEvent(EventId.TASK_START, taskId());
		
		// 全タスク起動まで待機
		waitForStartEvent(EventId.TASK_START, TaskId.ALL);
		
	_MAIN_LOOP:
		while (true)
		{
			try {
				startEvent(EventId.TASK_START, taskId);
				result = control();
			} catch (Exception e) {
				result = ReturnCode.FAILURE;		// ここにブレークポイント
				errorLog("control Excetpion: ".dup ~ e.toString);
			} finally {
				terminate();
			}
			endEvent(EventId.TASK_START, taskId);
			
			switch (result) {
			case ReturnCode.TERMINATE:
				log("Terminate");
				result = ReturnCode.SUCCESS;
				break _MAIN_LOOP;
			
			case ReturnCode.CONTINUE:
				log("Continue");
				try {
					initialize();
				} catch (Exception e) {
					errorLog("initialize Exception: ".dup ~ e.toString);
					result = ReturnCode.FAILURE;
					break _MAIN_LOOP;
				}
				continue _MAIN_LOOP;
				
			case ReturnCode.FAILURE:
				errorLog("Failure");
				break _MAIN_LOOP;
			// default: exception
			}
			break;
		}
		
		return cast(int)result;
	}
	
public:
	this(mailbox request_mail_id, mailbox response_mailbox,
						Environment environ, TaskResource resource)
	{
		m_response_mailbox 	= response_mailbox;
		s_environ			= environ;
		super(request_mail_id, resource);
	}
	
	Mail waitForRequest()
	{
		return receiveMail();
	}
	
	Mail fetchRequest()
	in {
		assert(pollMail());
	}
	body {
		return receiveMail();
	}
	
	bool emptyRequest()
	{
		return !pollMail();
	}
	
	bool pollRequest(uint wait)
	{
		return pollMail(wait);
	}
	
	void sendRequestTo(mailbox to, Mail request)
	{
//		request.to		= to;	// これいらんよな？
		request.from	= m_response_mailbox;
		resource.mailBox.send(to, request);
	}
	
	Mail waitForResponse()
	{
		return resource.mailBox.receive(m_response_mailbox);
	}
	
	Environment environ()
	{
		return s_environ;
	}
	
	Configuration config()
	{
		return s_environ.configuration;
	}
	
	Thread thread()
	{
		return m_thread;
	}
}

