﻿/**
 * Authors: http://d.hatena.ne.jp/ultraist/
 */
module pes.conv;
// to include 	std.conv.ConvError,
// 			and	std.conv.ConvOverflowError
 import std.conv;
private import std.string;
import std.stdio;
// 変換

char[] binaryToHexString(ubyte[] data)
{
	char[] hex = new char[data.length * 2];
	uint x;
	uint b;
	
	x = 0;
	for (int i = 0; i < data.length; ++i) {
		b = (data[i] & 0xf0) >> 4;
		hex[x++] = b > 0x09 ? ('a' + (b - 0x0a)):('0' + b);
		b = (data[i] & 0x0f);
		hex[x++] = b > 0x09 ? ('a' + (b - 0x0a)):('0' + b);
	}
	
	return hex;
}

ubyte[] hexStringToBinary(char[] hex)
in {
	assert(hex.length % 2 == 0);
}
body {
	ubyte[] bin = new ubyte[hex.length / 2];
	uint x;
	
	for (int i = 0; i < hex.length; i += 2) {
		bin[x++] = (hexToUByte(hex[i]) << 4) | hexToUByte(hex[i + 1]);
	}
	
	return bin;
}

ubyte hexToUByte(char hex)
{
	if ('0' <= hex && hex <= '9') {
		return cast(ubyte)hex - '0';
	} else if ('a' <= hex && hex <= 'f') {
		return cast(ubyte)(hex - 'a' + 10);
	} else if ('A' <= hex && hex <= 'F') {
		return cast(byte)(hex - 'a' + 10);
	}
	
	throw new ConvError("" ~ hex);
}



private enum Type: long {
	HEX,
	OCT,
	DEC
}

long stringToLong(char [] s)
{
	bool	sign 	= false;
	Type	type	= Type.DEC;
	long	lv, i;
	int		pos;
	
	if (s.length == 0) {
		return 0L;
	}
	if (s[0] == '+' || s[0] == '-') {
		sign = s[0] == '-' ? true: false;
		if (++pos == s.length) {
			return 0L;
		}
	}
	if (s[pos] == '0') {
		if (++pos == s.length) {
			return 0L;
		}
		if (s[pos] == 'x' || s[pos] == 'X') {
			type = Type.HEX;
			if (++pos == s.length) {
				throw new ConvError(s);
			}
		} else {
			type = Type.OCT;
		}
	}
	
	try {
		switch (type) {
		case Type.DEC:
			lv = decimalStringToLong(s[pos .. length], 10);
			break;
		case Type.OCT:
			lv = decimalStringToLong(s[pos .. length], 8);
			break;
		case Type.HEX:
			lv = hexStringToLong(s[pos .. length]);
			break;
		}
	} catch (ConvError) {
		throw new ConvError(s);
	} catch (ConvOverflowError) {
		throw new ConvOverflowError(s);
	} catch (Object e) {
		throw new Exception(e.toString);
	}
	
	return sign ? -lv: lv;
}


long decimalStringToLong(char [] s, int weight)
in {
	assert(0 < weight && weight <= 10);
}
body {
	long 	lv, iof;
	int		i;
	
	for (i = 0; i < s.length; ++i) {
		if (! ('0' <= s[i] && s[i] <= '0' + weight - 1)) {
			throw new ConvError(s);
		}
		lv = lv * weight + s[i] - '0';
		
		if (iof > lv || (iof == lv && lv != 0)) {
			throw new ConvOverflowError(s);
		}
		iof = lv;
	}
	
	return lv;
}


long hexStringToLong(char [] s)
{
	long 	lv, iof;
	int		i;
	
	for (i = 0; i < s.length; ++i) {
		if (s[i] <= '9' && '0' <= s[i]) {
			lv = lv * 16 + s[i] - '0';
		} else if (s[i] <= 'F' && 'A' <= s[i]) {
			lv = lv * 16 + s[i] - 'A' + 10;
		} else if ('a' <= s[i] && s[i] <= 'f') {
			lv = lv * 16 + s[i] - 'a' + 10;
		} else {
			throw new ConvError(s);
		}
		if (iof > lv || (iof == lv && lv != 0)) {
			throw new ConvOverflowError(s);
		}
		iof = lv;
	}
	
	return lv;
}


debug(pes_conv)
{
	import std.stdio;
	
	int main()
	{
		static char [][] test = [
			"a", "+B", "-c", "1", "-1", "+1", "1234", "-1234", "+1234",
			"0x", "0x1", "-0x1", "+0x1", "0x12aB", "-0x12aB", "+0x12aB",
			"00", "0a", "+01", "-0C", "01234", "-01234", "+01234", "04567", "089",
			"100000000000000000000000000000000", "0x11111111111111111", "0x1111111111111111"
		];
		
		for (int i = 0; i < test.length; ++i) {
			try {
				writef("%s to %s\n", test[i], stringToLong(test[i]));
			} catch (Object e) {
				writef("%s\n", e.toString);
			}
		}
		
		return 0;
	}
}


