/**
 * Authors: http://d.hatena.ne.jp/ultraist/
 */
module pes.net.winny.key;

// Winny キークラス

private import std.socket;
private import std.c.time;
private import pes.byteorder;
private import pes.prompt;
private import pes.checksum;
private import pes.rc4;
private import pes.net.winny.command;

debug(query_task)
{
	import pes.prompt;
}

class NyKeyInformation {
protected:
	struct KeyInfoH {
	align(1):
		ViaAddress	key_address;
		ViaAddress	bbs_address;
		uint		file_size;
		ubyte[16]	hash;
		ubyte		filename_length;
		ubyte[2]	checksum;
	}
//	char[] filename
	struct KeyInfoB {
	align(1):
		ubyte[11]	trip;
		ubyte		bbs_trip_length;
	}
//	ubyte[] bbs_auth
	struct KeyInfoF {
	align(1):
		ushort		timer;
		uint		ref_count;
		uint		unix_time;
		ubyte		ignore_flag;
		ubyte		key_version;
	}
	uint		m_file_name_mbs_length;
	uint		m_key_address;
	ushort		m_key_port;
	uint		m_bbs_address;
	ushort		m_bbs_port;
	uint		m_file_size;
	ubyte[]		m_hash;
	char[]		m_file_name;
	char[]		m_trip;
	ubyte[]		m_bbs_trip;
	ushort		m_timer;
	ushort		m_ref_block;
	time_t		m_mtime;
	ubyte		m_ignore;
	ubyte		m_key_version;
	
public:
	this(ubyte[] key_block)
	{
		unpack(key_block);
	}
	this() {}
	
	ubyte[] pack()
	{
		uint pos;
		ubyte[] data;
		ushort	sum;
		ubyte[] mbs_filename;
		KeyInfoH key_h;
		KeyInfoB key_b;
		KeyInfoF key_f;
		
		key_h.key_address.address 	= networkToHost32(m_key_address);
		key_h.key_address.port		= m_key_port;
		key_h.bbs_address.address 	= networkToHost32(m_bbs_address);
		key_h.bbs_address.port		= m_bbs_port;
		key_h.file_size				= m_file_size;
		key_h.hash[0 .. length]		= m_hash[0 .. length];
		
		mbs_filename 				= utf8ToMBS(m_file_name);
		sum 						= sum16(mbs_filename);
		key_h.filename_length		= mbs_filename.length;
		key_h.checksum[0]			= cast(ubyte)sum & 0xff;
		key_h.checksum[1]			= cast(ubyte)((sum & 0xff00) >> 8);
		
		// アロケート
		data = new ubyte[KeyInfoH.sizeof + mbs_filename.length + KeyInfoB.sizeof + m_bbs_trip.length + KeyInfoF.sizeof];
		
		data[0 .. KeyInfoH.sizeof] 				= (cast(ubyte *)&key_h)[0 .. KeyInfoH.sizeof];
		pos = KeyInfoH.sizeof;
		data[pos .. pos + mbs_filename.length] 	= pes.rc4.RC4.crypt(key_h.checksum[0 .. 1], mbs_filename);
		pos += mbs_filename.length;
		
		key_b.trip[0 ..m_trip.length]			= cast(ubyte[])m_trip[0 .. length];
		key_b.bbs_trip_length					= m_bbs_trip.length;
		
		data[pos .. pos + KeyInfoB.sizeof]		= (cast(ubyte *)&key_b)[0 .. KeyInfoB.sizeof];
		pos += KeyInfoB.sizeof;
		
		data[pos .. pos + m_bbs_trip.length] 	= m_bbs_trip;
		pos += m_bbs_trip.length;
		
		key_f.timer 				= m_timer;
		key_f.ref_count 			= m_ref_block;
		key_f.unix_time 			= m_mtime;
		key_f.ignore_flag	 		= m_ignore;
		key_f.key_version 			= m_key_version;
		
		data[pos .. length] = (cast(ubyte *)&key_f)[0 .. KeyInfoF.sizeof];
		
		return data;
	}
	
	void unpack(inout ubyte[] data)
	{
		ubyte[] decname;
		
		if (data.length < KeyInfoH.sizeof) {
			throw new NyBadCommand("Query: キーヘッダ");
		}
		KeyInfoH *key_h = cast(KeyInfoH *)data.ptr;
		data			= data[KeyInfoH.sizeof .. length];
		
		m_key_address			= hostToNetwork32(key_h.key_address.address);
		m_key_port				= key_h.key_address.port;
		m_bbs_address			= hostToNetwork32(key_h.bbs_address.address);
		m_bbs_port				= key_h.bbs_address.port;
		m_file_size				= key_h.file_size;
		m_hash					= key_h.hash.dup;
		m_file_name_mbs_length	= key_h.filename_length;
		
		if (data.length < key_h.filename_length) {
			throw new NyBadCommand("Query: ファイル名長");
		}
		
		ushort	sum = (key_h.checksum[1] << 8) | key_h.checksum[0];
//		if (checkSum16(sum, data[0 .. key_h.filename_length])) {
//			m_file_name = mbsToUTF8(data[0 .. key_h.filename_length]);
//		} else {
			decname = pes.rc4.RC4.crypt(key_h.checksum[0 .. 1],
											data[0 .. key_h.filename_length]);
			if (checkSum16(sum, decname)) {
				m_file_name = mbsToUTF8(decname.dup);
			} else {
				throw new NyBadCommand("NyKeyInfo: ファイル名復号不可能");
			}
//		}
		
		data = data[key_h.filename_length .. length];
		
		if (data.length < KeyInfoB.sizeof) {
			throw new NyBadCommand("NyKeyInfo: キーボディ");
		}
		KeyInfoB *key_b = cast(KeyInfoB *)data.ptr;
		data			= data[KeyInfoB.sizeof .. length];
		
		m_trip = getCString(key_b.trip);
		
		if (data.length < key_b.bbs_trip_length) {
			throw new NyBadCommand("NyKeyInfo: BBSトリップ認証情報");
		}
		m_bbs_trip	= data[0 .. key_b.bbs_trip_length].dup;
		data		= data[key_b.bbs_trip_length .. length];
		
		if (data.length < KeyInfoF.sizeof) {
			throw new NyBadCommand("NyKeyInfo: キーフッタ");
		}
		
		KeyInfoF *key_f = cast(KeyInfoF *)data.ptr;
		data			= data[KeyInfoF.sizeof .. length];
		m_timer			= key_f.timer;
		m_ref_block		= key_f.ref_count;
		m_mtime			= key_f.unix_time;
		m_ignore		= key_f.ignore_flag;
		m_key_version	= key_f.key_version;
	}
	
	uint length()
	{
//		return KeyInfoH.sizeof + m_file_name_mbs_length + KeyInfoB.sizeof + m_bbs_trip.length + KeyInfoF.sizeof;
		return KeyInfoH.sizeof + utf8ToMBS(m_file_name).length + KeyInfoB.sizeof + m_bbs_trip.length + KeyInfoF.sizeof;
	}
	
	char[] toString()
	{
		return 
			std.string.format(	"KEY ADDRESS:	%s BBS_ADDRESS: %s\n"
								"FILE SIZE:   	%d\n"
								"HASH: 			%s\n"
								"FILE NAME: 	%s\n"
								"TRIP: 			%s\n"
								"BBS TRIP: 		%s\n"
								"TIMER: 		%d\n"
								"REF BLOCK: 	%d\n"
								"MTIME: 		%s\n"
								"IGNORE: 		%x\n"
								"VERSION: 		%x\n",
				(new InternetAddress(m_key_address, m_key_port)).toString(),
				(new InternetAddress(m_bbs_address, m_bbs_port)).toString(),
				m_file_size, m_hash,
				m_file_name, m_trip, m_bbs_trip,
				m_timer, m_ref_block, m_mtime, m_ignore, m_key_version);
	}
	
	// get
	uint fileNameMBSLength()
	{
		return m_file_name_mbs_length;
	}
	
	uint keyAddress()
	{
		return m_key_address;
	}
	
	ushort keyPort()
	{
		return m_key_port;
	}
	
	InternetAddress address()
	{
		return new InternetAddress(m_key_address, m_key_port);
	}
	
	uint bbsAddress()
	{
		return m_bbs_address;
	}
	
	ushort bbsPort()
	{
		return m_bbs_port;
	}
	
	uint fileSize()
	{
		return m_file_size;
	}
	
	ubyte[] hash()
	{
		return m_hash;
	}
	
	char[] fileName()
	{
		return m_file_name;
	}
	
	char[] trip()
	{
		return m_trip;
	}
	
	ubyte[] bbsTrip()
	{
		return m_bbs_trip;
	}
	
	ushort timer()
	{
		return m_timer;
	}
	
	ushort referredBlock()
	{
		return m_ref_block;
	}
	
	time_t modifiedTime()
	{
		return m_mtime;
	}
	
	ubyte ignore()
	{
		return m_ignore;
	}
	
	ubyte keyVersion()
	{
		return m_key_version;
	}
	
	// setter
	
	void fileNameMBSLength(uint len)
	{
		m_file_name_mbs_length = len;
	}
	
	void keyAddress(uint address)
	{
		m_key_address = address;
	}
	
	void keyPort(ushort port)
	{
		m_key_port = port;
	}
	
	void bbsAddress(uint address)
	{
		m_bbs_address = address;
	}
	
	void bbsPort(ushort port)
	{
		m_bbs_port = port;
	}
	
	void fileSize(uint file_size)
	{
		m_file_size = file_size;
	}
	
	void hash(ubyte[] hash)
	{
		m_hash = hash;
	}
	
	void fileName(char[] filename)
	{
		m_file_name = filename;
	}
	
	void trip(char[] trip)
	{
		m_trip = trip.dup;
	}
	
	void bbsTrip(ubyte[] bbstrip)
	{
		m_bbs_trip = bbstrip;
	}
	
	void timer(ushort time)
	{
		m_timer = time;
	}
	
	void referredBlock(ushort ref_block)
	{
		m_ref_block = ref_block;
	}
	
	void modifiedTime(time_t mtime)
	{
		m_mtime = mtime;
	}
	
	void ignore(ubyte ignore_v)
	{
		m_ignore = ignore_v;
	}
	
	void keyVersion(ubyte key_version)
	{
		m_key_version = key_version;
	}
	
	bool tick()
	{
		if (m_timer == 0) {
			return false;
		}
		--m_timer;
		
		return true;
	}
	
	void tick10()
	{
		
	}
	
	void destroy()
	{
//		delete m_hash;
		delete m_file_name;
		delete m_bbs_trip;
		delete m_trip;
		m_hash = null;
		m_file_name = null;
		m_bbs_trip = null;
		m_trip = null;
	}
}

