﻿/**
 * Authors: http://d.hatena.ne.jp/ultraist/
 */
 
// NestedThreadMutex Win32


module pes.system.semaphore;
private	import	pes.system.all;

version(Win32)
{
	private import std.c.windows.windows;
	
	struct CRITICAL_SECTION {
		void *DebugInfo;
		
		LONG LockCount;
		LONG RecursionCount;
		HANDLE OwningThread;
		HANDLE LockSemaphore;
		ULONG *SpinCount;
		char[128] dummy;
	};
	extern(Windows) {
		BOOL InitializeCriticalSection(CRITICAL_SECTION *);
		BOOL InitializeCriticalSectionAndSpinCount(CRITICAL_SECTION *, DWORD);
		VOID EnterCriticalSection(CRITICAL_SECTION *);
		VOID LeaveCriticalSection(CRITICAL_SECTION *);
		BOOL TryEnterCriticalSection(CRITICAL_SECTION *);
		VOID DeleteCriticalSection(CRITICAL_SECTION *);
	}
} else {
	pragma(msg, "warning: pes.system.semaphore: not implemented for this compiler version");
	static assert(0);
}

public alias	uint	semaphore;

public SemaphoreResource	GlobalSemaphore;

static this()
{
	GlobalSemaphore = new SemaphoreResource;
}

private import std.stdio;

public class NestedThreadMutex
{
protected:
	CRITICAL_SECTION	m_cs;
	
public:
	this()
	{
		InitializeCriticalSection(&m_cs);
	}
	
	this(uint spin_count)
	{
		InitializeCriticalSectionAndSpinCount(&m_cs, spin_count);
	}
	
//	~this()
//	{
//		DeleteCriticalSection(&m_cs);
//	}
	
	void enter()
	{
		EnterCriticalSection(&m_cs);
	}
	
	void leave()
	{
		LeaveCriticalSection(&m_cs);
		Task.yield();
	}
	
	void  tryEnter()
	{
		if (!TryEnterCriticalSection(&m_cs)) {
			return false;
		}
		
		return true;
	}
}

class SemaphoreResource
{
protected:
	NestedThreadMutex[semaphore] 	m_semaphore;
	
public:
	this()
	{
	}
	
	void create(semaphore id)
	{
		m_semaphore[id] = new NestedThreadMutex();
	}
	
	void signal(semaphore id)
	{
		NestedThreadMutex mutex;
		
		synchronized(this) {
			mutex = m_semaphore[id];
		}
		mutex.enter();
	}
	
	void release(semaphore id)
	{
		NestedThreadMutex mutex;
		
		synchronized(this) {
			mutex = m_semaphore[id];
		}
		mutex.leave();
	}
	
	semaphore[] ids()
	{
		semaphore[] keys;
		
		synchronized(this) {
			keys = m_semaphore.keys;
		}
		
		return keys;
	}
}

debug(pes_system_semaphore)
{
	import	std.stdio;
	import	std.c.time;
	import	pes.system.all;
	
	enum: semaphore {
		SEM1,
	}
	
	class Data {
		int data;
		this (int i)
		{
			data = i;
		}
		override char[] toString()
		{
			return std.string.toString(data);
		}
	}
	
	void processing()
	{
		uint max = rand() % 10000;
		uint hoge;
		
		for (uint i = 0; i < max; ++i) {
			hoge *= cast(int)((i * 2.34) / 3.12);
		}
	}
	
	class Task1: Task
	{
		this(uint id, TaskResource resource)
		{
			super(id, resource);
		}
		
		int mainLoop()
		{
			
			for (int i = 0; i < 100000; ++i) {
//				delayTask(rand() % 10);
				signalSemaphore(SEM1);
				processing();
				if (i == 0) {
					msleep(1000);
				}
				processing();
				writef("task1: get, release %d\n", i);
				releaseSemaphore(SEM1);
			}
			
			return 0;
		}
		
	}
	
	class Task2: Task
	{
		this(uint id, TaskResource resource)
		{
			super(id, resource);
		}
		
		int mainLoop()
		{
			
			for (int i = 0; i < 100000; ++i) {
//				processing();
//				delayTask(rand() % 10);
				signalSemaphore(SEM1);
				processing();
				signalSemaphore(SEM1);
				processing();
				writef("task2: get, release %d\n", i);
				releaseSemaphore(SEM1);
				processing();
				releaseSemaphore(SEM1);
			}
			
			return 0;
		}
		
	}
	
	class Task3: Task
	{
		this(uint id, TaskResource resource)
		{
			super(id, resource);
		}
		
		int mainLoop()
		{
			
			for (int i = 0; i < 100000; ++i) {
//				processing();
//				delayTask(rand() % 10);
				signalSemaphore(SEM1);
				signalSemaphore(SEM1);
				processing();
				signalSemaphore(SEM1);
				signalSemaphore(SEM1);
				signalSemaphore(SEM1);
				signalSemaphore(SEM1);
				signalSemaphore(SEM1);
				signalSemaphore(SEM1);
				signalSemaphore(SEM1);
				processing();
				writef("task3: get, release %d\n", i);
				releaseSemaphore(SEM1);
				releaseSemaphore(SEM1);
				releaseSemaphore(SEM1);
				processing();
				releaseSemaphore(SEM1);
				releaseSemaphore(SEM1);
				releaseSemaphore(SEM1);
				releaseSemaphore(SEM1);
				releaseSemaphore(SEM1);
				releaseSemaphore(SEM1);
			}
			
			return 0;
		}
		
	}
	
	class Task4: Task
	{
		this(uint id, TaskResource resource)
		{
			super(id, resource);
		}
		
		int mainLoop()
		{
			
			for (int i = 0; i < 100000; ++i) {
//				delayTask(rand() % 10);
				signalSemaphore(SEM1);
				writef("task4: get, release %d!!\n", i);
				releaseSemaphore(SEM1);
			}
			
			return 0;
		}
		
	}
	
	import std.random;
	
	int main()
	{
		TaskResource	resource = new TaskResource;
		resource.semaphore.create(SEM1);
		resource.semaphore.create(1);
//		resource.semaphore.create(2);
//		resource.semaphore.create(3);
		
		Task t1 = new Task1(0, resource);
		Task t2 = new Task2(0, resource);
		Task t3 = new Task3(0, resource);
		Task t4 = new Task4(0, resource);
		
		t1.start();
		t2.start();
		t3.start();
		t4.start();
		
		
		t1.wait();
		t2.wait();
		t3.wait();
		
		t4.wait();
		
		
		return 0;
	}
}

