module pes.system.synchronizer;

// 同期オブジェクト Win32版

version(Win32)
{
} else {
	pragma(msg, "warning: pes.system.synchronizer: not implemented for this compiler version")
	static assert(0);
}

private import std.c.windows.windows;

extern(Windows)
{
	HANDLE CreateEventA(VOID *, BOOL, BOOL, LPCTSTR *);
	BOOL SetEvent(HANDLE);
	BOOL ResetEvent(HANDLE);
}
alias CreateEventA CreateEvent;
const DWORD		INFINITE		= 0xFFFFFFFF;
const DWORD		WAIT_OBJECT_0	= 0x00000000;
const DWORD		WAIT_TIMEOUT	= 0x00000102;


public class SynchronizerException:Exception
{
	this(char[] msg)
	{
		super(msg);
	}
}

public class Synchronizer
{
protected:
	HANDLE 	m_event;
	bool	m_closed;
	
public:
	this()
	{
		m_event	= CreateEvent(null, FALSE, FALSE, null);
		m_closed = false;
	}
	
	~this()
	{
		if (!m_closed) {
			CloseHandle(m_event);
			m_closed = true;
		}
	}
	
	void close()
	{
		if (!m_closed) {
			CloseHandle(m_event);
			m_closed = true;
		}
	}
	
	void wait()
	{
		wait(INFINITE);
	}
	
	bool wait(uint msec)
	{
		uint result;
		
		result = WaitForSingleObject(m_event, msec);
		if (result == WAIT_TIMEOUT) {
			return false;
		}
		
		if (result != WAIT_OBJECT_0) {
			throw new SynchronizerException("error wait");
		}
		
		return true;
	}
	
	void signal()
	{
		if (SetEvent(m_event) == 0) {
			throw new SynchronizerException("error signal");
		}
	}
	
	void reset()
	{
		if (ResetEvent(m_event) == 0) {
			throw new SynchronizerException("error reset");
		}
	}
}

