﻿/**
 * Authors: http://d.hatena.ne.jp/ultraist/
 */
module task.nodetask;

// ノード管理タスク
// リスナー

private import	std.thread,
				std.socket,
				std.stream,
				std.c.time,
				std.string,
				std.random;

private import	pes.system.all,
				pes.net.winny.all,
				pes.variant;

private import common.environment;
private import common.task;

private import task.uicommand;
private import task.timertask;

debug(ui_task)
{
	import std.stdio;
	import std.cstream;
	import pes.prompt;
	import pes.dump;
}
debug(query_task)
{
	import pes.prompt;
}

class NodeTask: TaskBase
{
protected:
	const uint CLOSE_TICK	= 0;
	const uint CLUSTER_TICK	= 1;
	const uint SAVE_TICK	= 2;
	
	ConnectionTable		m_upstream_connection;	// 上流検索リンク
	ConnectionTable		m_downstream_connection;// 下流検索リンク
	ConnectionTable		m_ul_data_connection;	// データ転送リンク
	NodeTable	 		m_node_table;			// ノードテーブル
	
	NyConnection		m_listener;				// リスナー
	
	
	override TaskId taskId()
	{
		return TaskId.NODE;
	}
	
	override void initialize()
	{
		gcLimit = gcLimit / 2;
		m_listener = new NyConnection;
		m_listener.timeout(config.timeout, 0);
		m_listener.listen(config.fileSharingPort);
		debug(node_task) mbswritef(derr, "Node: リスナー起動(port:%d)\n", config.fileSharingPort);
		
		loadNodeAddress();
	}
	
	
	override void terminate()
	{
		// リンクタスクをとめる
		startEvent(EventId.RUNNING, FunctionId.LINK_INTERRUPT);
		waitForEndEvent(EventId.RUNNING, FunctionId.LINK_RUN);
		
		// ノードリスト保存
		saveNodeAddress();
		
		m_listener.close();
	}
	
	
	override ReturnCode control()
	{
		ReturnCode result;
		uint close_count, save_count, clustering_count;
		
		/* リスナータスクメインループ */
		while (true) {
			result = control_x(close_count, save_count, clustering_count);
			if (result != ReturnCode.SUCCESS) {
				break;
			}
			delayTask(10);
		}
		
		return result;
	}
	
	ReturnCode control_x(inout uint close_count,
							inout uint save_count,
							inout uint clustering_count)
	{
		Mail mail;
		
		updateNode();
			
		if (! pollRequest(50)) {
			try {
				/* ソケットをポーリング */
				listen();
				
				/* 検索リンクを構築 */
//				if (m_upstream_connection.length < config.upstreamConnectionMax + 1
				if (m_upstream_connection.length < config.upstreamConnectionMax
					&& !pollStartEvent(EventId.RUNNING, FunctionId.LINK_RUN))
				{
					sendRequestTo(MailId.LINK_REQUEST, Mail.makeRequest(MessageId.LINK_CONNECT, null));
				}
			} catch(Exception e) {
				debug(node_task) mbswritef(derr, "Node: Exception %s\n", e.toString);
			}
		} else {
			
			mail = fetchRequest();
			
			switch (mail.messageId) {
			case MessageId.TERMINATE:
				return ReturnCode.TERMINATE;//  終了
				
			case MessageId.REBOOT:
				if (rebootTest()) {
					sendRequestTo(MailId.UI_RESPONSE,
								Mail.makeResponse(MessageId.REBOOT, RequestStatus.OK, null));
					debug(ui_task) mbswritef(derr, "Node: リブート\n");
				} else {
					sendRequestTo(MailId.UI_RESPONSE,
								Mail.makeResponse(MessageId.REBOOT, RequestStatus.NG, null));
					break;
				}
				
				return ReturnCode.CONTINUE;
				
			case MessageId.REFRESH_NODE:
				try {
					saveNodeAddress();
					loadNodeAddress();
				} catch (Exception e) {
					log(format("例外: %s\n", e.toString));
				}
				break;
				
			case MessageId.CLOCK_TICK:
				switch ((cast(Variant)mail.data).toLong) {
				case CLOSE_TICK:
					// 上流検索リンクが最大数+1以上で、50秒立ったら1回線切断する
					if (m_upstream_connection.length >= config.upstreamConnectionMax + 1) {
						// クラスタリング欲求が先、それでも超えているなら
						desireForFellowFeeling();
						if (m_upstream_connection.length >= config.upstreamConnectionMax + 1) {
							desireForClose();
						}
					}
					break;
					
				case CLUSTER_TICK:
					// 60秒に一度コネクションをチェック
					// 同方向最大1コネクションを切断する
					if (m_downstream_connection.length != 0
						|| m_upstream_connection.length != 0) 
					{
							desireForFellowFeeling();
					}
					break;
				
				case SAVE_TICK:
				// 1分に一度noderefをセーブする
					try {
						saveNodeAddress();
					} catch (Exception e) {
						log(format("例外: %s\n", e.toString));
					}
					break;
				default:
					assert(0);
				}
				break;
				
			case MessageId.ADD_NODE:
				addNode(cast(NodeDetails)mail.data);
				break;
			default:
				break;
			}
			fullCollect();
		}
		
		return ReturnCode.SUCCESS;
	}
	
	// 切断欲求
	void desireForClose()
	{
		time_t		min_time 	= 0;
		time_t		cur_time	= time(null);
		
		Connection 	connection;
		
		m_upstream_connection.signal();
		try {
			try {
				debug(node_task) mbswritef(derr, "-- Node: 切断欲求\n");
				
				foreach (Connection e; m_upstream_connection.values) {
					if (min_time < e.connectTime) {
						min_time	= e.connectTime;
						connection	= e;
					}
				}
				if (cur_time - min_time > 50) {
					debug(node_task) mbswritef(derr, "Node: 切断(%s)\n",
						connection.nodeDetails.inetAddress.toString);
					
					m_upstream_connection.remove(connection.aaKey); // + close
				}
			} catch (SocketException) {
				// nothing
			}
		} finally {
			m_upstream_connection.release();
		}
	}
	
	// クラスタリング欲求
	// 切断の間隔はわりとテキトウなので注意
	void desireForFellowFeeling()
	{
		time_t 	cur_time		= time(null);
		uint	cluster_min		= uint.max;
		Connection	connection;
		
		// 一番短いクラスタ
		foreach (char[] key; config.clusteringKey) {
			if (cluster_min > key.length) {
				if (key.length != 0) {
					cluster_min = key.length;
				}
			}
		}
		
		if (cluster_min == uint.max) {
			cluster_min = 0; // < 条件は満たされない
		} else {
			cluster_min /= 2; // 1/2
		}
		
		debug(node_task) mbswritef(derr, "Node: クラスタリング欲求\n");
		m_upstream_connection.signal();
		try {
			foreach (Connection e; m_upstream_connection.values) {
				connection = null;
				if (cur_time - e.connectTime > 900) {
					// 15分以上経過
					connection = e;
				} else if (e.nodeDetails.speed == config.speed
					&& cur_time - e.connectTime > 300)
				{
					// 速度関係があまりよくなく5分経過
					connection = e;
				} else if (e.nodeDetails.speed < config.speed
					&& cur_time - e.connectTime > 30) 
				{
					// 速度関係がよくなく30秒経過
					connection = e;
				} else if (e.nodeDetails.correlation < cluster_min
					&& cur_time - e.connectTime > 50)
				{
					// 相関性があまりなく50秒以上経過
					connection = e;
				}
				if (connection !is null) {
					m_upstream_connection.remove(connection.aaKey);
					break;
				}
			}
		} finally {
			m_upstream_connection.release();
		}
		
		m_downstream_connection.signal();
		try {
			foreach (Connection e; m_downstream_connection.values) {
				connection = null;
				if (cur_time - e.connectTime > 900) {
					// 15分以上経過
					connection = e;
				} else if (e.nodeDetails.speed == config.speed
					&& cur_time - e.connectTime > 300)
				{
					// 速度関係があまりよくなく5分経過
					connection = e;
				} else if (e.nodeDetails.speed > config.speed
					&& cur_time - e.connectTime > 30)
				{
					// 速度関係がよくなく30秒経過
					connection = e;
				} else if (e.nodeDetails.correlation < cluster_min
					&& cur_time - e.connectTime > 50) 
				{
					// 相関性があまりなく50秒以上経過
					connection = e;
				}
				if (connection !is null) {
					m_downstream_connection.remove(connection.aaKey);
					break;
				}
			}
		} finally {
			m_downstream_connection.release();
		}
		
		m_ul_data_connection.signal();
		try {
			foreach (Connection e; m_ul_data_connection.values) {
				connection = null;
				if (!e.isLock
					&& cur_time - e.connectTime > 30
					&& e.connection.rate == 0) 
				{
					// 転送していないデータコネクション
					connection = e;
				}
				if (connection !is null) {
					m_ul_data_connection.remove(connection.aaKey);
				}
			}
		} finally {
			m_ul_data_connection.release();
		}
	}
	
	// リブートテスト
	bool rebootTest()
	{
		NyConnection test_listener 	= new NyConnection;
		bool test_ok;
		
		debug(ui_task) mbswritef(derr, "Node: リブートテスト(port:%s)\n", config.fileSharingPort);
		
		try {
			test_listener.listen(config.fileSharingPort);
			test_ok = true;
		} catch (Exception e) {
			test_ok = false;
		} finally {
			test_listener.close();
		}
		debug(ui_task) mbswritef(derr, "Node: リブートテスト %s\n", test_ok ? "OK":"NG");
		
		return test_ok;
	}
	
	
	// 接続者を調べる
	void listen()
	{
		NyProtocolHeader 	header;
		NyConnection 		client;
		NyConnectionType	type;
		Connection			connection;
		NodeDetails			node;
		
		try {
			m_listener.poll(NyConnection.Poll.READABLE);
			m_listener.timeout(config.timeout, 0);	// クライアント側タイムアウト変更
			client = m_listener.accept(header); 	// 接続認証
			
			connection = new Connection;
			node = connection.accept(config, client, header, type);
			
		} catch (SocketException) {
			/* 新規接続者なし */
			//delete connection;
			//delete header;
			if (connection is null && client !is null) {
				client.close();
			} else if (connection !is null) {
				connection.close();
			}
			return;
		}
		finally {
			m_listener.timeout(0, 10);
		}
		
		/* バージョンが違うのは怖いので切断する	*/
//		if (!(node.major == NyProtocolHeader.WINNY_MAJOR 
//			&& node.minor == NyProtocolHeader.WINNY_MINOR))
		if (node.minor != NyProtocolHeader.WINNY_MINOR)
		{
			try {
				connection.close(new NyLowVersion);
			} catch (SocketException) {
				// nothing
			}
			
			return;
		}
		
		/* 自分からの接続は切断する	*/
		if (node.isSelf(config))
		{
			try {
				connection.close();
			} catch (SocketException) {
				// do nothing
			}
			
			return;
		}
		
		/* 今のところポート0未対応なので切断する	*/
		if (node.isPort0) {
			try {
				connection.close(new NyReject);
			} catch (SocketException) {
				// do nothing
			}
			
			return;
		}
		
		/* 接続テスト		*/
		if (type.isBadPort0Connection) {
			try {
				connection.close();
			} catch (SocketException) {
				// do nothing
			}
			
			return;
		}
		
		/* 新規接続あり */
		
		// 接続テスト
		if (node.type != NodeDetails.NodeType.PORT0) {
			debug(node_task) mbswritef(derr, "Node: 接続テスト(%s)\n",node.inetAddress.toString);
			
			if (!connectTest(node.inetAddress)) {
				// テストNG
				try {
					connection.close(new NyWrongListeningPort);
				} catch (SocketException) {
					// nothing
				}
				debug(node_task) mbswritef(derr, "Node: 接続テストNG\n");
				return;
			} else {
				debug(node_task) mbswritef(derr, "Node: 接続テストOK\n");
			}
		}
		
		if (type.isTransferLink) {
			/* ファイル転送リンク */
			if (m_ul_data_connection.length < config.dataConnectionMax
				&& !m_ul_data_connection.exists(connection.aaKey))
			{
				// uploader/testでかぶる可能性があるので
				// 接続元ポート番号を使う
				connection.nodeDetails.inetAddress = connection.connection.address;
				
				m_ul_data_connection.set(connection.aaKey, connection);
				debug(node_task) 
					mbswritef(derr, "Node: 新しい 転送リンク(%s:%s:%s)\n",
										node.inetAddress.toString, header.major, header.minor);
			} else {
				/* 接続数が多い、または既に接続がある場合は切断する	*/
				NyConnectedLimitation close = new NyConnectedLimitation;
				try {
					connection.connection.sendCommand(close);
				} catch (SocketException ) {
				
				}
				finally {
					connection.close();
				}
			}
			
			return ; // RETURN
		}
		// ノード情報をテーブルに追加する
		addNode(node);
		
		// ノード情報を渡す
		try {
			connectedLimitation(connection, 5);
		} catch (SocketException ) {
			connection.close();
			return;
		}
		
		// 下流検索リンク
		if (m_downstream_connection.length >= config.downstreamConnectionMax
			|| connection.nodeDetails.speed > config.speed
			|| m_downstream_connection.exists(node.aaKey))
		{
			/* 接続数が多いか、自局より速度が速いか、既に接続がある場合は切断する	*/
			try {
				connectedLimitation(connection, 20);
			} catch (SocketException) {
				/// nothing
			} finally {
				connection.close();
			}
			
			return ;
		}
		if (!connection.nodeDetails.isDownstreamNode(config.speed)
			&& connection.nodeDetails.speed < config.speed) 
		{
			// 遅すぎるノードからの接続
			if (rand() % 3) {// 2/3
				try {
					connectedLimitation(connection, 20);
				} catch (SocketException) {
					/// nothing
				} finally {
					connection.close();
				}
				return;
			}
		}
		
		m_downstream_connection.set(connection.aaKey, connection);
		
		debug(node_task) 
			mbswritef(derr, "Node: 新しい 検索リンク(下流)(%s)\n",
								connection.nodeDetails.inetAddress);
	}
	
	// 同時接続数制限 のときは他ノード情報を送ったあとに切断する
	void connectedLimitation(Connection connection, uint max)
	{
		NyConnectedLimitation	close	= new NyConnectedLimitation;
		NyAnotherNode			oth		= new NyAnotherNode;
		NodeDetails[]			node;
		
		node = m_node_table.searchUpstreamNode(connection.nodeDetails, max);
		
		foreach (NodeDetails e; node) {
			oth.address 		= e.inetAddress.addr;
			oth.fileSharingPort	= e.inetAddress.port;
			oth.speed			= e.speed;
			oth.clusteringKey	= e.clusteringKey;
			
			connection.connection.sendCommand(oth);
		}
	}
	
	// 接続テスト
	bool connectTest(InternetAddress address)
	{
		Connection connection = new Connection;
		bool result;
		
		result = connection.connectTest(config, address);
		
		return result;
	}
	
	// 他タスクからのノード追加要求処理
	// 追加する確率はわりとテキトウなので注意
	void addNode(NodeDetails node)
	in {
		assert(node !is null);
	}
	body {
		debug(node_task) mbswritef(derr, "Node: ノード情報追加(%s)\n", node.inetAddress.toString);
		
		// 
		// 上流検索リンクになりえる場合
		if (node.isUpstreamNode(config.speed)) {
			// ノードテーブルへ追加
			node.priorityUp(NodeDetails.FluctuationLevel.MIDDLE);
			m_node_table.set(node.aaKey, node);
		}
		// 保持ノード数が限度の20%以下でノードとの相関性がある場合
		else if (config.nodeMax * 0.20 > m_node_table.length 
					&& node.speed > config.speed 
					&& node.correlation > 0
					&& rand() % 2 == 0)
		{
			// 一時的に優先順位を上げる
			// 一度繋いだ後は、リンクタスク側で優先順をさらに下げる
			node.priorityUp(NodeDetails.FluctuationLevel.MIDDLE);
			// ノードテーブルへ追加
			m_node_table.set(node.aaKey, node);
		} else {
			if (node.speed > config.speed) {
				// 一定の確率で速いノードを混ぜる・・
				if (rand() % 8 == 0) {
					node.priorityUp(NodeDetails.FluctuationLevel.MIDDLE);
				} else {
					node.priorityUp(NodeDetails.FluctuationLevel.LOW);
				}
			} else {
				node.priorityDown(NodeDetails.FluctuationLevel.HI);
				node.priorityDown(NodeDetails.FluctuationLevel.HI);
			}
			m_node_table.set(node.aaKey, node);
		}
		node = null;
	}
	
	// 更新の必要があればノードテーブルを更新する
	void updateNode()
	{
		if (m_node_table.length < config.nodeMax) {
			return; // なし
		}
		
		m_node_table.signal();
		try {
			NodeDetails[] node = m_node_table.values.sort;
			node = node[0 .. length - cast(int)(config.nodeMax * 0.60)]; // 40%に削る
			
			m_node_table.clear();
			
			foreach (NodeDetails e; node) {
				m_node_table.set(e.aaKey, e);
			}
			m_node_table.rehash();
		} finally {
			m_node_table.release();
		}
		
		saveNodeAddress();
	}
	
	
	// noderefへノード保存
	void saveNodeAddress()
	{
		uint max = cast(uint)(config.nodeMax * 0.10);
		InternetAddress[] savelist = new InternetAddress[max];
		uint i;
		
		m_node_table.rehash;
		foreach (NodeDetails e; m_node_table.values.sort) {
			savelist[i++] = e.inetAddress;
			if (i >= max) {
				break;
			}
		}
		if (i != max) {
			savelist = savelist[0 .. i];
		}
		
		try {
			NyNodeAddress.save(config.noderef, savelist);
		} catch {
			// nothing
		}
		
		savelist = null;
	}
	
	
	// noderefからノードを読み込み
	void loadNodeAddress()
	{
		InternetAddress[] addrlist = NyNodeAddress.load(config.noderef);
		
		m_node_table.clear();
		
		foreach (InternetAddress e; addrlist) {
			NodeDetails node = new NodeDetails(e);
			m_node_table.set(node.aaKey, node);
			
			if (m_node_table.length > config.nodeMax * 0.08) {
				break;
			}
		}
		debug(node_task) mbswritef(derr, "Node: ノードリスト読み込み(%d件)\n", m_node_table.length);
		
		addrlist = null;
	}
	
public:
	this(Environment environ, TaskResource resource)
	{
		super(MailId.NODE_REQUEST,
				MailId.NODE_RESPONSE, environ, resource);
		
		// ローカルに取っとく
		m_node_table 			= environ.nodeTable;
		m_upstream_connection	= environ.upstreamConnection;
		m_downstream_connection	= environ.downstreamConnection;
		m_ul_data_connection	= environ.ulDataConnection;
		
		// 切断タイマー 30sec
		setMessageTimer(this, 30, MessageTimer.TTL.ANY, new VLong(CLOSE_TICK));
		
		// クラスタ監視タイマー 20sec
		setMessageTimer(this, 20, MessageTimer.TTL.ANY, new VLong(CLUSTER_TICK));
		
		// ノードテーブル保存タイマー 60sec
		setMessageTimer(this, 60, MessageTimer.TTL.ANY, new VLong(SAVE_TICK));
	}
}

