﻿/**
 * Authors: http://d.hatena.ne.jp/ultraist/
 */
module task.timertask;

// タイマータスク (常駐)

// 割り込みはメッセージ受信イベントで行うので、
// 受け取り方は自由
// タイマーを登録すると
// CLOCK_TICKメッセージを送信する

private import 	std.thread,
				std.c.time,
				std.socket,
				std.random;

private import 	pes.system.all,
				pes.queue;

private import 	common.task,
				common.environment;
private import  pes.prompt;

class MessageTimer
{
public:
	enum TTL
	{
		ANY,
		ONCE,
	}
	
	mailbox receiver;
	Object	data;
	uint	interval;
	TTL		ttl;
	
	this(mailbox id, uint intsec, TTL ltyp, Object arg)
	{
		receiver 	= id;
		interval	= intsec;
		ttl			= ltyp;
		data		= arg;
	}
}

private class MessageTimeCounter
{
	MessageTimer 	timer;				// timer args
	uint			decrementCounter;	// sec
	Object			aaKey;				// aaKey
	
	this(MessageTimer timer)
	{
		this.timer = timer;
		
		decrementCounter 	= timer.interval;
		aaKey				= this;
	}
}


public void 
setMessageTimer(TaskBase task, uint interval_sec,
					MessageTimer.TTL ttl, Object data = null)
{
	MessageTimer timer =
			new MessageTimer(task.mailboxId, interval_sec, ttl, data);
	
	task.sendRequestTo(MailId.TIMER_REQUEST, 
				Mail.makeRequest(MessageId.SET_TIMER, timer));
}


class TimerTask: TaskBase
{
protected:
	MessageTimeCounter[Object] m_counter;
	
	override TaskId taskId()
	{
		return TaskId.TIMER;
	}
	
	override void initialize()
	{
		// priorityを下げる
		setPriority(PRIORITY.IDLE);
		gcLimit = gcLimit * 2;
	}
	
	override void terminate()
	{
		//
	}
	
	override ReturnCode control()
	{
		ReturnCode result = ReturnCode.SUCCESS;
		Mail message;
		MessageTimeCounter counter;
		
		time_t last_time = 0;
		time_t current_time;
		
	MAIN_LOOP:
		while (true) {
			current_time = time(null);
			if (current_time != last_time) {
				// 1sec経過
				last_time = current_time;
				
				foreach (MessageTimeCounter e; m_counter.values) {
					if (--e.decrementCounter == 0) {
						// send tick 
						sendRequestTo(e.timer.receiver, 
									Mail.makeRequest(MessageId.CLOCK_TICK, e.timer.data));
						
						if (e.timer.ttl == MessageTimer.TTL.ONCE) {
							m_counter.remove(e.aaKey);
						} else {
							e.decrementCounter = e.timer.interval;
						}
					}
				}
			}
			
			fullCollect();
			// メッセージ受信
			if (pollRequest(500)) {
				message = fetchRequest();
				switch (message.messageId) {
				case MessageId.TERMINATE:
					result = ReturnCode.TERMINATE;
					break MAIN_LOOP;
					
				case MessageId.SET_TIMER:
					counter = new MessageTimeCounter(cast(MessageTimer)message.data);
					m_counter[counter.aaKey] = counter;
					break;
				default:
					break;
				}
			}
			
		}
		
		return result;
	}
	
public:
	this(Environment environ, TaskResource resource)
	{
		super(MailId.TIMER_REQUEST,
				MailId.TIMER_RESPONSE, environ, resource);
	}
}
