﻿module common.keytable;

private import std.string;
private import std.utf;
private import std.stdio;

private import pes.wstring;
private import pes.net.winny.key;
private import pes.system.locker;
private import pes.system.semaphore;
import std.stdio;
// キーテーブルのキー(構造体はmemcmp的な比較)
// MD5はコリジョンを人為的に発生可能なので一応サイズまで入れとく
class KeyTableKey
{
	ubyte[]		m_hash;
	uint		m_size;
	uint		m_magic;
	
	this(ubyte[] data_hash, uint data_size)
	{
		m_hash	= data_hash.dup;
		m_size	= data_size;
		
		m_magic = hash();
	}
	
	uint hash()
	{
		uint magic = m_size;
		foreach (ubyte e; m_hash) {
			magic = (magic * 13) + e;
		}
		
		return magic;
	}
	
	override uint toHash()
	{
		return m_magic;
	}
	
	int opCmp(Object o)
	{
		KeyTableKey key = cast(KeyTableKey)o;
		
		int diff = m_size - key.m_size;
		
		if (diff == 0) {
			return std.string.cmp(cast(char[])m_hash, cast(char[])key.m_hash);
		}
		
		return diff;
	}
	
	char[] toString()
	{
		return format("hash(%s): toHash(%s): hash: %s , size:%s", m_magic, hash, m_hash, m_size);
	}
}


// キー要素
class KeyInformation
{
protected:
	KeyTableKey			m_aa_key;
	NyKeyInformation	m_key;
	
public:
	NyKeyInformation key()
	{
		return m_key;
	}
	
	// ソート用
	override int opCmp(Object o)
	{
		KeyInformation k = cast(KeyInformation)o;
		
		return k.m_key.timer - m_key.timer;
	}
	
	this(NyKeyInformation k)
	{
		m_key 		= k;
		m_aa_key	= new KeyTableKey(k.hash, k.fileSize);
	}
	
	KeyTableKey aaKey()
	{
		return m_aa_key;
	}
}

class KeyTable: Locker!(KeyTableKey, KeyInformation)
{
public:
	this(SemaphoreResource resource, semaphore id)
	{
		super(resource, id);
	}
	
	void addKeyArray(NyKeyInformation[] array, LockerMode mode = LockerMode.AUTO)
	{
		KeyInformation key;
		Object lock;
		
		if (mode == LockerMode.AUTO) {
			lock = lockObject;
		} else {
			lock = new Object;
		}
		
		synchronized(lock) {
			foreach (NyKeyInformation e; array) {
				key = new KeyInformation(e);
				set(key.aaKey, key, LockerMode.NON);
			}
		}
	}
	
	void close(){}
	
	NyKeyInformation[] searchHash(ubyte[] hash, uint max = uint.max, LockerMode mode = LockerMode.AUTO)
	{
		NyKeyInformation[] result;
		
		Object lock;
		
		if (mode == LockerMode.AUTO) {
			lock = lockObject;
		} else {
			lock = new Object;
		}
		synchronized(lock) {
			foreach(KeyInformation e; values(LockerMode.NON)) {
				if (e.key.hash == hash) {
					result ~= e.key;
				}
				if (max-- == 0) {
					break;
				}
			}
		}
		
		return result;
	}
	
	NyKeyInformation[] searchName(char[] keyword, uint max = uint.max, LockerMode mode = LockerMode.AUTO)
	{
		NyKeyInformation[] result;
		
		Object lock;
		
		if (mode == LockerMode.AUTO) {
			lock = lockObject;
		} else {
			lock = new Object;
		}
		
		synchronized(lock) {
			foreach(KeyInformation e; values(LockerMode.NON)) {
				if (findKeyword(toUTF16(e.key.fileName), toUTF16(keyword))) {
					result ~= e.key;
				}
				if (max-- == 0) {
					break;
				}
			}
		}
		
		return result;
	}
	
	NyKeyInformation[] searchTrip(char[] trip, uint max = uint.max, LockerMode mode = LockerMode.AUTO)
	{
		NyKeyInformation[] result;
		Object lock;
		
		if (mode == LockerMode.AUTO) {
			lock = lockObject;
		} else {
			lock = new Object;
		}
		
		synchronized(lock) {
			foreach(KeyInformation e; values(LockerMode.NON)) {
				if (e.key.trip == trip) {
					result ~= e.key;
				}
				if (max-- == 0) {
					break;
				}
			}
		}
		
		return result;
	}
	void clear(uint clear_count, LockerMode mode = LockerMode.AUTO)
	{}
	
	void clear(LockerMode mode = LockerMode.AUTO) 
	{
		super.clear(mode);
	}
	
	void tick(LockerMode mode = LockerMode.AUTO)
	{
		Object lock;
		
		if (mode == LockerMode.AUTO) {
			lock = lockObject;
		} else {
			lock = new Object;
		}
		synchronized(lock) {
			foreach (KeyInformation e; values(LockerMode.NON)) {
				if (!e.key.tick()) {
					remove(e.aaKey);
				}
			}
		}
	}

}
